import numpy as np
import sklearn
import cvxopt
import sklearn.metrics.pairwise
from sklearn.ensemble import IsolationForest
from functools import reduce
import aad

class nuSVMaad(object):
    """Semisupervised Outlier Detection
    
    Parameters
    ----------
    kernel : {'rbf','if'}, default='rbf'
         Specifies the kernel type to be used in the algorithm.
         It must be one of 'rbf', 'if'.
         If none is given, 'rbf' will be used.
    gamma : rbf kernel parameter, default=1.
        Tolerance for stopping criterion.
    nu1 : float, default=0.5
        An upper bound on the fraction of training
        errors and a lower bound of the fraction of support
        vectors for labelled data. Should be in the interval (0, 1]. By default 0.5
        will be taken.
    nu2 : float, default=0.5
        An upper bound on the fraction of training
        errors and a lower bound of the fraction of support
        vectors for unlabelled data. Should be in the interval (0, 1]. By default 0.5
        will be taken.
    Attributes
    ----------
    alpha : Lagrange multipliers to define the decision function.
    sv: Support vectors.
    ind_sv : Index of X corresponding to support vectors.
    Y_sv : Labels (+/-1) of support vectors (+1 if corresponding to unlabelled data).
    alpha_trunc : Lagrange multipliers truncated to ensure the sparsity of the support vectors.
    sv_trunc : Corresponding truncated support vectors.
    ind_sv_trunc : Ind of X corresponding to truncated support vectors.
    Y_sv_trunc : Labels (+/-1) of truncated support vectors (+1 if corresponding to unlabeled data).
    rho1 : Margin for labeled vectors.
    rho2 : Margin for unlabeled vectors.
    b : Decision function bias.
    """
    def __init__(self, kernel='rbf', gamma=1, nu1=0.5, nu2=0.5, unsupervised_ad = None, t=1):

        if nu1<0 or nu1>1:
            raise Exception("nu should be between 0 and 1")
        if nu2<0 or nu2>1:
            raise Exception("nu should be between 0 and 1")
            
        self.kernel = kernel
        self.gamma = gamma
        self.t = t
        self.nu1 = nu1
        self.nu2 = nu2
        
    def fit_unsupervised(self,X):
        if self.kernel == 'rbf':
            unsupervised_ad = sklearn.svm.OneClassSVM(nu = self.nu2, kernel = 'rbf', gamma = self.gamma)
            unsupervised_ad.fit(X)
        elif self.kernel == 'IF':
            [n,d] = X.shape
            phi = min(256,n)
            unsupervised_ad = IsolationForest(max_samples = phi, n_estimators = self.t)
            unsupervised_ad.fit(X)
            
        self.unsupervised_ad = unsupervised_ad

    def fit(self, X, y, tol = 1e-5, verbose = True):
        """Detects the soft boundary of the set of samples X.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples_labelled + n_samples_unlabelled, n_features)
            Set of samples, where n_samples_labelled is the number of labelled samples, n_samples_unlabelled
            is the number of unlabelled samples, and n_features is the number of features.
        y : {array-like} of shape (n_samples_labelled). Set of labels for labelled vectors.
        Returns
        tol : Tolerance to trunc support vectors.
        verbose : Boolean. If True, displays optimization steps.
        """
        nu = self.nu1

        if np.any(np.abs(y)!=1):
            raise Exception("The vectors of labels should only contains +1 (nominal) or -1 (anomaly)")

        [r,d] = X.shape
        
        #Check nu1 value
        cp1 = sum(y == 1)
        cm1 = sum(y == -1)
        nu1max = 2*min([cp1, cm1])/r
        if(nu > nu1max):
            nu = 0.999*nu1max
            print("Warning: nu1 was superior the the limit and was thus set to",nu)
            
        Y = np.diag(y.ravel())

        # Check kernel
        if(self.kernel == 'rbf'):            
            G = sklearn.metrics.pairwise.rbf_kernel(X, gamma = self.gamma)
            
            alpha_0 = self.unsupervised_ad.dual_coef_.transpose()
            
            phiphi0 = sklearn.metrics.pairwise.rbf_kernel(X, self.unsupervised_ad.support_vectors_, gamma = self.gamma)
            G0 = sklearn.metrics.pairwise.rbf_kernel(self.unsupervised_ad.support_vectors_, gamma = self.gamma)

            norm_w0 = np.sqrt(alpha_0.transpose().dot(G0.dot(alpha_0)))
            
            q = Y.dot(phiphi0.dot(alpha_0))/norm_w0
            
            
        elif(self.kernel == 'none'):
            G = np.dot(X,X.transpose())
        elif(self.kernel == 'IF'):  
            H_data = []
            for i_tree in range(self.unsupervised_ad.n_estimators):
                h_data, weight_tree = aad.path_length(X, self.unsupervised_ad.estimators_[i_tree])
                H_data.append(weight_tree)  # faster to append list
                
            phi_data = np.concatenate(H_data, axis=1)
            
            w0 = np.ones([phi_data.shape[1],1])
            norm_w0 = np.linalg.norm(w0)
            wp = w0/norm_w0
            self.wp = wp.ravel()
            
            G = phi_data.dot(phi_data.transpose())
            q = Y.dot(phi_data.dot(wp))                 
        else:
            raise Exception("Unknown kernel")
        
        self.norm_w0 = norm_w0
        
        # Compute Optimization problem matrices and constraints
        #Y = np.diag(np.append(y,np.ones(n-r)))
        P = Y.dot(G).dot(Y)
        

        Gc = np.vstack((-np.eye(r), np.eye(r), -np.ones([1,r])))
        hc = np.vstack((np.zeros([r,1]), (1/r)*np.ones([r,1]), -nu))

        Ac = y.reshape([1,r]).astype(np.double)
        bc = np.zeros([1,1])
        
        alpha = cvxopt_solve_qp(P = P, q = q, G = Gc, h = hc, A = Ac, b = bc, verbose = verbose) # solve the dual

        # Get support vectors
        sv = np.where(alpha>0)[0]
        sv_trunc = np.where(alpha>tol)[0]
        
        # Truncate alphas value
        alpha_trunc=np.zeros(r)
        alpha_trunc[sv_trunc] = alpha[sv_trunc]
        alpha_trunc[np.where(abs(alpha-1/r)<tol)] = 1/r
        
        # Compute score inside the constrains
        #score_training_data =  self.t/norm_w0*self.unsupervised_ad.score_samples(X).ravel()# + G.dot(Y.dot(alpha))
        if self.kernel == 'IF':
            score_training_data =  phi_data.dot(self.wp) + G.dot(Y.dot(alpha))
        elif self.kernel == 'rbf':
            score_training_data =  (1/norm_w0*self.unsupervised_ad.score_samples(X)).ravel() + G.dot(Y.dot(alpha)) 

        I_rho_m_b = reduce(np.intersect1d, (np.where(alpha>tol),np.where(abs(alpha-1/r)>tol), np.where(y.ravel() == 1)))
        I_rho_p_b = reduce(np.intersect1d, (np.where(alpha>tol),np.where(abs(alpha-1/r)>tol), np.where(y.ravel() == -1)))
        
        if I_rho_m_b.size == 0:
            rho_m_b = score_training_data[np.where(y.ravel() == 1)].max()
        else:
            rho_m_b = np.mean(score_training_data[I_rho_m_b])
            
        if I_rho_p_b.size == 0:
            rho_p_b = (-score_training_data[np.where(y.ravel() == -1)]).max()
        else:
            rho_p_b = np.mean(-score_training_data[I_rho_p_b])
        
        # Compute margin and bias
        b = (rho_p_b-rho_m_b)/2
        rho = (rho_p_b+rho_m_b)/2
        
        # Save results
        self.alpha = alpha
        self.sv = X[sv]
        self.ind_sv = sv
        self.alpha_trunc = alpha_trunc[sv_trunc]
        self.ind_sv_trunc = sv_trunc
        self.sv_trunc = X[sv_trunc]
        self.Y_sv_trunc = Y[sv_trunc,:][:,sv_trunc]
        self.Y_sv = Y
        self.rho = rho
        self.b = b
        self.scores_mem = score_training_data
        if(self.kernel == 'IF'):
            self.w_trunc = Y.dot(alpha)[sv_trunc].transpose().dot(phi_data[sv_trunc])
            self.w = Y.dot(alpha)[sv].transpose().dot(phi_data[sv])
            self.phi_mat = phi_data
        
        return self
    
    def score_samples(self, X2):
        """Computes the scores of the test vectors in X2 using the truncated support vectors.
        Parameters
        ----------
        X2 : {array-like, sparse matrix} of shape (n_samples_X2, n_features)
            Set of samples, where n_samples_X2 is the number of samples and
            n_features is the number of features.
        Returns
        -------
        scores_X2 : {array-like, sparse matrix} of shape (n_samples_X2, 1)
            Scores of each samples in X2.
        """
        if(self.kernel == 'rbf'):
            scores_X2 = sklearn.metrics.pairwise.rbf_kernel(X2, self.sv_trunc, gamma = \
                       self.gamma).dot(self.Y_sv_trunc.dot(self.alpha_trunc)) + (1/self.norm_w0)*self.unsupervised_ad.score_samples(X2) 
        elif(self.kernel == 'IF'):
            H_X2 = []
            for i_tree in range(self.unsupervised_ad.n_estimators):
                h_data, weight_tree = aad.path_length(X2, self.unsupervised_ad.estimators_[i_tree])
                H_X2.append(weight_tree)  # faster to append list
                
            phi_X2 = np.concatenate(H_X2, axis=1)
            phi_sv = self.phi_mat
            
            G = phi_X2.dot(phi_sv.transpose())
            
            #scores_X2 = phi_X2.dot(self.wp) 
            scores_X2 = phi_X2.dot(self.w+self.wp) 
            #scores_X2 = (1/self.norm_w0*self.unsupervised_ad.score_samples(X2)).ravel()*self.t + G.dot(self.Y_sv.dot(self.alpha))
        else:
            raise Exception("Unknown kernel")   
        
        return scores_X2.ravel()
    
    def decision_function(self, X2):
        """Computes the scores of the test vectors in X2 using the truncated support vectors.
        Parameters
        ----------
        X2 : {array-like, sparse matrix} of shape (n_samples_X2, n_features)
            Set of samples, where n_samples_X2 is the number of samples and
            n_features is the number of features.
        Returns
        -------
        scores_X2 : {array-like, sparse matrix} of shape (n_samples_X2, 1)
            Scores of each samples in X2.
        """         
        
        return (self.score_samples(X2) + self.b).ravel()
    
    def predict(self, X2):
        """Predict the label of data in X2 using truncated support vectors.
        Parameters
        ----------
        X2 : {array-like, sparse matrix} of shape (n_samples, n_features)
            Set of samples to get prediction from.
        Returns
        pred_X2 : Prediction of labels.
        """
        labels_X2 = np.sign(self.decision_function(X2))
        
        return labels_X2
       
    
def cvxopt_solve_qp(P, q, G=None, h=None, A=None, b=None, init_val = None,verbose = True):
    """Solves quadratic constrained optimization problem using Quadratic Programming.
    Paramaters
    ----------
    Solves argmin_x (1/2)*x^T*P*x +q^T*x s.t. G*x<=h and A*x=b
    Returns
    -------
    alpha : Solution of optimization problem.
    """
    P = .5 * (P + P.T)  # make sure P is symmetric
    args = [cvxopt.matrix(P), cvxopt.matrix(q)]
    if G is not None:
        args.extend([cvxopt.matrix(G), cvxopt.matrix(h)])
        if A is not None:
            args.extend([cvxopt.matrix(A), cvxopt.matrix(b)])
            if init_val is not None:
                args.extend([None])
                args.extendend([init_val])
    cvxopt.solvers.options['show_progress'] = verbose
    sol = cvxopt.solvers.qp(*args)
    #if 'optimal' not in sol['status']:
    #    return None
    return np.array(sol['x']).reshape((P.shape[1],))
